/*
  uim-agent.c - backend program of uim.el
 
  Copyright (c) 2005, Konosuke Watanabe <nosuke@csc.ne.jp>
  All rights reserved.

  $Id: uim-agent.c,v 1.25 2005/07/24 15:42:52 nosuke Exp $

  Redistribution and use in source and binary forms, with or
  without modification, are permitted provided that the
  following conditions are met:

  1. Redistributions of source code must retain the above
     copyright notice, this list of conditions and the
     following disclaimer.
  2. Redistributions in binary form must reproduce the above
     copyright notice, this list of conditions and the
     following disclaimer in the documentation and/or other
     materials provided with the distribution.
  3. Neither the name of authors nor the names of its
     contributors may be used to endorse or promote products
     derived from this software without specific prior written
     permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/


#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <locale.h>

#include <sys/select.h>

#include <uim/uim.h>
#include <uim/uim-helper.h>
#include <uim/uim-util.h>
#include <uim/uim-im-switcher.h>

/* #define UIM_AGENT_DEBUG */

#define PREEDIT_SIZE 1024

int debug = 0;
FILE *debugf = NULL;


typedef struct uim_key{
  int mod;
  int key;
} uim_key;



typedef struct candidate_param {
  int valid;
  int num;
  int disp_limit;
  int index;
  char **candstr;
} candidate_param;



typedef struct preedit_buffer {
  char *str;
  int attr;
  unsigned cursor;
  struct preedit_buffer *next;
} preedit_buffer;

typedef struct uim_agent_context {
  uim_context context;
  int context_id;
  char *encoding;
  char *im;
  int preedit_length;
  preedit_buffer *preedit_head;
  preedit_buffer *preedit_tail;
  candidate_param *cand;
} uim_agent_context;


typedef struct im_encoding {
  char *im;
  char *encoding;
  struct im_encoding *next;
} im_encoding;


im_encoding *im_enc_list_head = NULL, *im_enc_list_tail = NULL;



char *default_engine_name = NULL;

/* current focused context */
uim_agent_context *current;

typedef struct uim_agent_context_list {
  uim_agent_context *agent_context;
  struct uim_agent_context_list *next;
  struct uim_agent_context_list *prev;
}uim_agent_context_list;

uim_agent_context_list *agent_context_list_head = NULL;
uim_agent_context_list *agent_context_list_tail = NULL;



void convert_keyname(char *keyname, const char *name);
enum UKey matchkey( const char *keyname );
int a_putchar( int c );
int a_printf( const char *fmt, ... );
int m_printf( int level, const char *fmt, ... );
void commit_cb( void *ptr, const char *str );
void clear_preedit( uim_agent_context *ua );
void preedit_clear_cb( void *ptr );
void preedit_pushback_cb( void *ptr, int attr, const char * str );
int show_preedit( uim_agent_context *ua );
void show_candidate( uim_agent_context *ua );
void preedit_update_cb( void *ptr );
void candidate_activate_cb( void *ptr, int num, int limit );
void candidate_select_cb( void *ptr, int index );
void candidate_shift_page_cb( void *ptr, int direction );
void clear_candidate( uim_agent_context *ua );
void candidate_deactivate_cb( void *ptr );
void prop_list_update_cb( void *ptr , const char *str );
void prop_label_update_cb( void *ptr , const char *str );
int check_im_name( const char *imname );
int list_im_engine( void );

im_encoding *search_im_encoding( const char *im );
im_encoding *new_im_encoding( const char *im, const char *encoding );
int set_im_encoding( const char *im, const char *encoding );
const char *get_im_encoding( const char *im );

uim_context create_context( const char *encoding, uim_agent_context *ptr );

uim_agent_context * create_uim_agent_context( const char *encoding );
uim_agent_context * new_uim_agent_context( int id, const char *encoding );
int release_uim_agent_context( int id );
uim_agent_context * get_uim_agent_context( int id );
uim_agent_context * switch_im( uim_agent_context *ua, const char *im, const char *encoding );
void helper_send_im_list( void );
void helper_im_changed( char *request, char *engine_name );
int focused( uim_agent_context *ua );
int unfocused( void );
void helper_handler( void );
void helper_disconnected_cb( void );
void check_helper_connection( void );
static int cmd_release( int serial, int context_id );
static int cmd_unfocused( int serial, int context_id );
static int cmd_focused( int serial, int context_id );
static int cmd_hide( int serial, int context_id );
static int cmd_show( int serial, int context_id );
static int cmd_new( int serial, int context_id, const char *encoding );
static int cmd_reset( int serial, int context_id );
static int cmd_change( int serial, int context_id, const char *im, const char *encoding );
static int cmd_list( int serial );
static int cmd_default( int serial, const char *im );
static int cmd_setenc( int serial, const char *im, const char *encoding );
static int cmd_prop( int serial, int context_id, const char *prop );
static int cmd_error( int serial );
int process_command( int serial, int cid, char *cmd );
int process_keyvector( int serial, int cid, uim_key ukey, const char *keyname );
int analyze_keyvector( char *vector, uim_key *ukey, char *keyname );
void wait_until_data_arrival( fd_set *rfds );
int take_one_line_from_buffer( char *buf, char *input );






void
convert_keyname( char *keyname, const char *name ){
  if( strcmp(name, "BS") == 0 ){
	strcpy(keyname, "backspace");
  }else if( strcmp(name, "TAB") == 0 ){
	strcpy(keyname, "tab");
  }else if( strcmp(name, "RET") == 0 ){
	strcpy(keyname, "return");
  }else if( strcmp(name, "ESC") == 0 ){
	strcpy(keyname, "escape");
  }else if( strcmp(name, "DEL") == 0 ){
	strcpy(keyname, "delete");
  }else if( strcmp(name, "DEL") == 0 ){
	strcpy(keyname, "delete");
  }else if( strcmp(name, "SPC") == 0 ){
	strcpy(keyname, "space");
  }else{
	strcpy(keyname, name);
  }
}

enum UKey
matchkey( const char *keyname ){
  if( !strcmp("backspace", keyname) )
	return UKey_Backspace;
  else if( !strcmp("tab", keyname) )
	return UKey_Tab;
  else if( !strcmp("return", keyname) )
	return UKey_Return;
  else if( !strcmp("escape", keyname) )
	return UKey_Escape;
  else if( !strcmp("prior", keyname) )
	return UKey_Prior;
  else if( !strcmp("next", keyname) )
	return UKey_Next;
  else if( !strcmp("end", keyname) )
	return UKey_End;
  else if( !strcmp("home", keyname) )
	return UKey_Home;
  else if( !strcmp("left", keyname) )
	return UKey_Left;
  else if( !strcmp("up", keyname) )
	return UKey_Up;
  else if( !strcmp("right", keyname) )
	return UKey_Right;
  else if( !strcmp("down", keyname) )
	return UKey_Down;
  else if( !strcmp("insert", keyname) )
	return UKey_Insert;
  else if( !strcmp("delete", keyname) )
	return UKey_Delete;
  else if( !strcmp("muhenkan", keyname) )
	return UKey_Muhenkan;
  else if( !strcmp("henkan", keyname) )
	return UKey_Henkan_Mode;
  else if( !strcmp("zenkaku-hankaku", keyname) )
	return UKey_Zenkaku_Hankaku;
  else if( !strcmp("f1", keyname) )
	return UKey_F1;
  else if( !strcmp("f2", keyname) )
	return UKey_F2;
  else if( !strcmp("f3", keyname) )
	return UKey_F3;
  else if( !strcmp("f4", keyname) )
	return UKey_F4;
  else if( !strcmp("f5", keyname) )
	return UKey_F5;
  else if( !strcmp("f6", keyname) )
	return UKey_F6;
  else if( !strcmp("f7", keyname) )
	return UKey_F7;
  else if( !strcmp("f8", keyname) )
	return UKey_F8;
  else if( !strcmp("f9", keyname) )
	return UKey_F9;
  else if( !strcmp("f10", keyname) )
	return UKey_F10;
  else if( !strcmp("f11", keyname) )
	return UKey_F11;
  else if( !strcmp("f12", keyname) )
	return UKey_F12;
  else if( !strcmp("f13", keyname) )
	return UKey_F13;
  else if( !strcmp("f14", keyname) )
	return UKey_F14;
  else if( !strcmp("f15", keyname) )
	return UKey_F15;
  else if( !strcmp("f16", keyname) )
	return UKey_F16;
  else if( !strcmp("f17", keyname) )
	return UKey_F17;
  else if( !strcmp("f18", keyname) )
	return UKey_F18;
  else if( !strcmp("f19", keyname) )
	return UKey_F19;
  else if( !strcmp("f20", keyname) )
	return UKey_F20;
  else if( !strcmp("f21", keyname) )
	return UKey_F21;
  else if( !strcmp("f22", keyname) )
	return UKey_F22;
  else if( !strcmp("f23", keyname) )
	return UKey_F23;
  else if( !strcmp("f24", keyname) )
	return UKey_F24;
  else if( !strcmp("f25", keyname) )
	return UKey_F25;
  else if( !strcmp("f26", keyname) )
	return UKey_F26;
  else if( !strcmp("f27", keyname) )
	return UKey_F27;
  else if( !strcmp("f28", keyname) )
	return UKey_F28;
  else if( !strcmp("f29", keyname) )
	return UKey_F29;
  else if( !strcmp("f30", keyname) )
	return UKey_F30;
  else if( !strcmp("f31", keyname) )
	return UKey_F31;
  else if( !strcmp("f32", keyname) )
	return UKey_F32;
  else if( !strcmp("f33", keyname) )
	return UKey_F33;
  else if( !strcmp("f34", keyname) )
	return UKey_F34;
  else if( !strcmp("f35", keyname) )
	return UKey_F35;
  else if( !strcmp("space", keyname) )
	return ' ';
  else
	return UKey_Other;
}



int helper_fd = -1;


int output_enable = 1;


int
a_putchar( const int c ){
  if( output_enable ){
	return putchar(c);
  }else{
	return 0;
  }
}


int
a_printf( const char *fmt, ... ){
  int ret = 0;
  
  if( output_enable ){
	va_list ap;

	va_start(ap, fmt);
	ret = vfprintf(stdout, fmt, ap);
	va_end(ap);  
  }

  return ret;
}


int
m_printf( int level, const char *fmt, ... ){
  int ret = 0;
  va_list ap;

  if( debug >= level ){
	va_start(ap, fmt);
	ret = vfprintf(stderr, fmt, ap);
	va_end(ap);  
  }else{
#ifdef UIM_AGENT_DEBUG
	if( debugf ){
	  va_start(ap, fmt);
	  ret = vfprintf(debugf, fmt, ap);
	  va_end(ap);
	}
#endif
  }
  return ret;
}



void
commit_cb( void *ptr, const char *str ){

  const char *p;

  m_printf(1, "<CB> commit_cb\n");


  a_printf("( s ");

  if( strchr(str, '"') ||  strchr(str, '\\') ) {
	p = str;
	a_putchar('"');
	while( *p ){
	  if( *p == '"' ){
		a_printf("\\\"");
	  }else if( *p == '\\' ){
		a_printf("\\\\");
	  }else{
		a_putchar(*p);
	  }
	  p++;
	}
	a_putchar('"');
  }else{
	a_printf("\"%s\"", str);
  }

  a_printf(" ) ");

}


void
clear_preedit( uim_agent_context *ua ){

  preedit_buffer *p, *ptmp;

  p = ua->preedit_head;

  while( p ){
	ptmp = p;
	p = p->next;
	free( ptmp->str );
	free( ptmp );
  }

  ua->preedit_head = ua->preedit_tail = NULL;
  ua->preedit_length = 0;

}


void
preedit_clear_cb( void *ptr ){

  uim_agent_context *ua = (uim_agent_context *)ptr;

  /* preedit_buffer *p, *ptmp; */

  m_printf(1, "<CB> preedit_clear_cb\n");

  clear_preedit( ua );  

  /*
  p = ua->preedit_head;

  while( p ){
	ptmp = p;
	p = p->next;
	free( ptmp->str );
	free( ptmp );
  }

  ua->preedit_head = ua->preedit_tail = NULL;
  ua->preedit_length = 0;
  */

}


void
preedit_pushback_cb( void *ptr, int attr, const char * str ){

  uim_agent_context *ua = (uim_agent_context *)ptr;

  preedit_buffer *p;

  m_printf(1, "<CB> preedit_pushback_cb \"%s\" (attr: %d)\n", str, attr); 

  p = (preedit_buffer *)malloc(sizeof(preedit_buffer));

  if( ua->preedit_head == NULL ){
	ua->preedit_head = p;
	ua->preedit_tail = p;
  }else{
	ua->preedit_tail->next = p;
	ua->preedit_tail = p;
  }


  if( strlen(str) > 0 ){
	p->str = (char *)malloc(strlen(str) + 1);
	strcpy(p->str, str);
	ua->preedit_length += strlen(str);
  }else{
	p->str = NULL;
  }
  p->attr = attr;
  p->next = NULL;

}



int
show_preedit( uim_agent_context *ua ){

  unsigned int i;

  preedit_buffer *p;

  p = ua->preedit_head;

  if( p == NULL || ua->preedit_length == 0 ){
	a_printf(" ( e ) ");
	return 0;
  }

  a_printf("( p ");

  while( p ){
	a_printf(" ( ");
	if( p->attr & UPreeditAttr_Reverse) a_putchar('r');
	if( p->attr & UPreeditAttr_UnderLine) a_putchar('u');
	if( p->attr & UPreeditAttr_Cursor) a_putchar('c');
	if( p->attr & UPreeditAttr_Separator) a_putchar('s');
	a_putchar('t');

	a_putchar(' ');

	a_putchar('"');
	if( p->str != NULL ){
	  if( !strchr(p->str, '"') && !strchr(p->str, '\\') ){
		a_printf("%s", p->str);
	  }else{
		for( i = 0; i < strlen(p->str); i ++ ){
		  if( p->str[i] == '\\' ){
			a_printf("\\\\");
		  }else if( p->str[i] == '"' ){
			a_printf("\\\"");
		  }else{
			a_putchar(p->str[i]);
		  }
		}
	  }
	}
	a_putchar('"');
	
	a_printf(" ) ");

	p = p->next;
  }
  a_printf(" ) ");

  return 1;
}



void
show_candidate( uim_agent_context *ua ){

  int i;
  int page;
  int index;
  const char *p;
  /* uim_candidate candidate; */

  if( ! ua->cand->valid || ua->cand->num == 0 ) return;

  index = ua->cand->index;
  page = index / ua->cand->disp_limit;


  a_printf("( c ");

  a_printf(" ( %d . %d ) ",  ua->cand->index + 1,  ua->cand->num );

  for( i = ua->cand->disp_limit * page; i < ua->cand->disp_limit * (page + 1); i ++ ){
	if( i >= ua->cand->num ) break;

	/* candidate = uim_get_candidate(ua->context, i, i); */

	if( i == index ) a_printf("( t . ");
	else a_printf("( nil . ");

	/* p = uim_candidate_get_cand_str(candidate); */
	p = ua->cand->candstr[i];

	if( strchr(p, '"')
		|| strchr(p, '\\') ){
	  
	  a_putchar('"');
	  while( *p ){
		if( *p == '"' ){
		  a_printf("\\\"");
		}else if( *p == '\\' ){
		  a_printf("\\\\");
		}else{
		  a_putchar(*p);
		}
		p++;
	  }
	  a_putchar('"');
	}else{
	  a_printf("\"%s\"", p);
	}
	
	/* uim_candidate_free(candidate); */
	
	a_printf(" ) ");
  }

  a_printf(") ");

  /* m_printf(1, "\n"); */


}



void
preedit_update_cb( void *ptr ){

  uim_agent_context *ua = (uim_agent_context *)ptr;

  m_printf(1, "<CB> preedit_update_cb (ua = %p)\n", ua);

  /* show_preedit(ua); */

  /* m_printf(1, "\n"); */
  
}


void
candidate_activate_cb( void *ptr, int num, int limit ){

  /* 
	 num:   total number of candidates
	 limit: candidates per page
  */
  int i;
  uim_candidate candidate;

  uim_agent_context *ua = (uim_agent_context *)ptr;

  m_printf(1, "<CB> candidate_activate_cb (num=%d,limit=%d)\n", num, limit);

  ua->cand->valid = 1;

  ua->cand->index = 0; 
  ua->cand->disp_limit = limit;
  ua->cand->num = num;

  ua->cand->candstr = (char **)malloc(sizeof(char *) * num);

  for( i = 0; i < num; i ++ ){
	candidate = uim_get_candidate(ua->context, i, limit ? i % limit : i);
	ua->cand->candstr[i] = strdup(uim_candidate_get_cand_str(candidate));
	uim_candidate_free(candidate);
  }

  /* show_candidate(ua); */

}



void
candidate_select_cb( void *ptr, int index ){

  uim_agent_context *ua = (uim_agent_context *)ptr;

  m_printf(1, "<CB> candidate_select_cb (index: %d)\n", index);

  ua->cand->index = index;

  /* show_candidate(ua); */
}


void
candidate_shift_page_cb( void *ptr, int direction ){

  int index;
  int is_first = 0, is_last = 0;

  uim_agent_context *ua = (uim_agent_context *)ptr;

  /*
	direction: 1 -> forward
	           0 -> backward
  */
  m_printf(1, "<CB> candidate_shift_page_cb (direction: %d)\n", direction);

  index = ua->cand->index;
  
  is_first = (index < ua->cand->disp_limit) ? 1 : 0;

  /*
   * ((cand.num - 1) / cand.disp_limit) + 1   : total pages
   * (index / cand.disp_limit) + 1            : current page number
   */
  is_last = ((((ua->cand->num - 1) / ua->cand->disp_limit) + 1) 
			 == ((index / ua->cand->disp_limit) + 1)) ? 1 : 0;
  
  if( direction == 0 ){  /* backward */
	if( is_first ){
      /* loop */
	  m_printf(2, "first page\n");
	  index = (ua->cand->num / ua->cand->disp_limit) * ua->cand->disp_limit + index;
	}else{
	  index -= ua->cand->disp_limit;
	}
  }else if( direction == 1 ){  /* forward */
	if( is_last ){
	  /* loop */
	  m_printf(2, "last page\n");
	  index = index % ua->cand->disp_limit;
	}else{
	  index += ua->cand->disp_limit;
	}
  }

  if( index < 0 ) index = 0;
  
  if( index >= ua->cand->num ) index = ua->cand->num - 1;

  ua->cand->index = index;

  uim_set_candidate_index(ua->context, index);
}


void
clear_candidate( uim_agent_context *ua ){
  int i;

  if( ua->cand->valid ){
	ua->cand->valid = 0;

	for( i = 0; i < ua->cand->num; i ++ ){
	  free(ua->cand->candstr[i]);
	}
	free(ua->cand->candstr);
  }

}


void
candidate_deactivate_cb( void *ptr ){

  /* int i; */
  uim_agent_context *ua = (uim_agent_context *)ptr;

  m_printf(1, "<CB> candidate_deactivate_cb\n");

  clear_candidate(ua);
  /*
  if( ua->cand->valid ){
	ua->cand->valid = 0;

	for( i = 0; i < ua->cand->num; i ++ ){
	  free(ua->cand->candstr[i]);
	}
	free(ua->cand->candstr);
  }
  */
}



void
prop_list_update_cb( void *ptr , const char *str ){
  
  unsigned len;
  char buf[4096];
  char *p[4];

  uim_agent_context *ua = (uim_agent_context *)ptr;

  m_printf(1, "<CB> prop_list_update_cb\n");
  
  if( ua != current ){
	m_printf(1, "<WARNING>ua != current\n");
	return;
  }

#define  PROP_LIST_FORMAT  "prop_list_update\ncharset=%s\n%s"  

  len = strlen( PROP_LIST_FORMAT) + strlen(ua->encoding) +
	strlen(str) + 1 ;
  
  if( len > sizeof(buf) ){
	m_printf(1, "<WARNING> buffer size over\n");
	return;
  }

  snprintf( buf, sizeof(buf), PROP_LIST_FORMAT, ua->encoding, str );

  uim_helper_send_message(helper_fd, buf) ;


  m_printf(2, "<DEBUG>  prop_list_update_cb \"%s\"\n", buf); 

  if( default_engine_name ){
	a_printf(" ( l ( \"%s\" . \"%s\" ) ", ua->im, default_engine_name );
  }else{
	a_printf(" ( l ( \"%s\" . \"%s\" ) ", ua->im, 
			 uim_get_default_im_name(setlocale(LC_ALL, NULL)));
  }

  strcpy(buf, str);
  p[0] = buf;

  while( p[0] && *p[0] ){
	p[1] = strchr(p[0], '\n');

	if( p[1] ){
	  *p[1] = '\0';
	}else{
	  break;
	}

	if( strlen(p[0]) >= 6 && strncmp(p[0], "branch", 6) == 0 ){
	  if( (p[2] = strchr(p[0], '\t')) && (p[3] = strchr(p[2] + 1, '\t')) ){
		*p[2] = *p[3] = '\0';
		a_printf(" ( %s %s ) ",p[2] + 1, p[3] + 1);
	  }
	}
	p[0] = p[1] + 1;
  }

  a_printf(" ) ");

}



void
prop_label_update_cb( void *ptr , const char *str ){
  
  unsigned len;
  char buf[4096];
  /*  char *p1, *p2; */

  uim_agent_context *ua = (uim_agent_context *)ptr;

  m_printf(1, "<CB> prop_label_update_cb\n");
  
  if( ua != current ){
	m_printf(1, "<WARNING> ua != current\n");
	return;
  }

#define PROP_LABEL_FORMAT "prop_label_update\ncharset=%s\n%s"

  len = strlen( PROP_LABEL_FORMAT) + strlen(ua->encoding)
	+ strlen(str) + 1 ;
  
  if( len > sizeof(buf) ){
	m_printf(1, "<WARNING> buffer size over\n");
	return;
  }

  snprintf( buf, sizeof(buf), PROP_LABEL_FORMAT, ua->encoding, str );

  uim_helper_send_message(helper_fd, buf) ;

  m_printf(2, "<DEBUG>  prop_label_update_cb \"%s\"\n", buf); 

  strcpy(buf, str);
  /*  p1 = buf; */

}



int
check_im_name( const char *imname ){

  int ret = 0, i;
  const char *name;
  uim_context context;

  context = uim_create_context( NULL, "UTF-8", NULL, NULL, NULL, NULL);
  for(i = 0 ; i < uim_get_nr_im(context); i++){  
	name = uim_get_im_name(context, i);
	if( strcmp(name, imname) == 0 ) ret = 1;
  }

  uim_release_context(context);

  return ret;
}



/* search encoding entry */
im_encoding *
search_im_encoding( const char *im ){

  im_encoding *im_enc = im_enc_list_head;

  m_printf(1, "<FN> search_im_encoding\n");

  while( im_enc ){
	if( strcmp(im_enc->im, im) == 0 ){
	  m_printf(1, "<FN>  found\n");
	  return im_enc;
	}
	im_enc = im_enc->next;
  }
  m_printf(1, "<FN>  not found\n");
  return NULL;
}


im_encoding *
new_im_encoding( const char *im, const char *encoding ){

  im_encoding *im_enc = (im_encoding *)malloc(sizeof(im_encoding));

  m_printf(1, "<FN> new_im_encoding\n");

  im_enc->im = strdup( im );

  if( encoding != NULL ){
	im_enc->encoding = strdup( encoding );
  }else{
	im_enc->encoding = NULL;
  }

  im_enc->next = NULL;

  if( im_enc_list_tail == NULL ){
	im_enc_list_tail = im_enc;
	im_enc_list_head = im_enc;
	m_printf(1, "<FN>  create as head\n");
  }else{
	im_enc_list_tail->next = im_enc;
	im_enc_list_tail = im_enc;
	m_printf(1, "<FN>  append to tail\n");
  }

  return im_enc;
}


/* set engine - enc list */
int
set_im_encoding( const char *im, const char *encoding ){

  im_encoding *im_enc;

  if( im == NULL ) return -1;


  im_enc = search_im_encoding(im);

  if( im_enc != NULL ){
	if( im_enc->encoding ) free( im_enc->encoding );

	if( encoding ) im_enc->encoding = strdup( encoding );
	else im_enc->encoding = NULL;

  }else{
	new_im_encoding( im, encoding );
  }
  return 1;
}



/* get encoding from im list */
const char *
get_im_encoding( const char *im ){

  im_encoding *im_enc = search_im_encoding(im);

  m_printf(1, "<FN> get_im_encoding\n");

  if( im_enc != NULL ){
	m_printf(1, "<FN> encoding = %s\n", im_enc->encoding);
	return im_enc->encoding;
  }else{
	return NULL;
  }
}



/* show supported IM engines */
int
list_im_engine( void ){

  int i;

  uim_context context;

  context = uim_create_context( NULL, "UTF-8", NULL, NULL, NULL, NULL);

  if( default_engine_name ){
	a_printf(" ( L \"%s\" ", default_engine_name);
  }else{
	a_printf(" ( L \"%s\" ", uim_get_default_im_name(setlocale(LC_ALL,NULL)));
  }
  
  for(i = 0 ; i < uim_get_nr_im(context); i++){  
	const char *name, *lang, *language, *shortd, *encoding;
	name = uim_get_im_name(context, i);

	lang = uim_get_im_language(context, i);

	language = uim_get_language_name_from_locale(lang);

	shortd = uim_get_im_short_desc(context, i);
	encoding = uim_get_im_encoding(context, i);

	a_printf(" ( \"%s\" \"%s\" \"%s\" \"%s\" %s ) ", name, lang, language, shortd, encoding);
	
  }
  a_printf(" ) ");

  uim_release_context(context);

  return i;
}



uim_context
create_context( const char *encoding, uim_agent_context *ptr ){

  uim_context context;

  context = uim_create_context( ptr,
								encoding,
								NULL,
								default_engine_name,
								uim_iconv,
								commit_cb );

  uim_set_preedit_cb( context,
					  preedit_clear_cb,
					  preedit_pushback_cb,
					  preedit_update_cb );

  uim_set_candidate_selector_cb( context,
								 candidate_activate_cb,
								 candidate_select_cb,
								 candidate_shift_page_cb,
								 candidate_deactivate_cb );

  uim_set_prop_list_update_cb( context,
							   prop_list_update_cb );

  uim_set_prop_label_update_cb( context,
								prop_label_update_cb );
  
  return context;

}





uim_agent_context *
create_uim_agent_context( const char *encoding ){

  uim_agent_context *ret;
  const char *im;

  m_printf(2, "<FN> create_uim_agent_context\n");

  /* get default engine name at first time */
  /* if( ! default_engine_name ){ */

	/* if valid_default_im_name does not exist,
	   uim_get_default_im_name returns "direct" */
  /*	default_engine_name 
	  = strdup(uim_get_default_im_name(setlocale(LC_ALL,NULL)));

	m_printf(2, "<MSG> set default IM: %s\n", default_engine_name);
	}*/


  ret = (uim_agent_context *)malloc(sizeof(uim_agent_context));

  if( encoding ){
	ret->encoding = strdup(encoding);
  }else{
	if( debug ){
	  ret->encoding = strdup("EUC-JP");
	}else{
	  ret->encoding = strdup("UTF-8");
	}
  }


  ret->context = create_context(ret->encoding, ret);

  /* 
	 uim_get_current_im_name doesn't return IM name.
	 It just returns copy of argument.
	 Hmm... I just want to know IM name which is in use.
  */
  /*
  if( (im = uim_get_current_im_name(ret->context)) ){
	ret->im = strdup(im);
  }else{
	ret->im = NULL;
  }
  */

/*if( (im = uim_get_default_im_name(ret->encoding)) ){ */
  if( (im = uim_get_default_im_name(setlocale(LC_ALL, NULL))) ){
	ret->im = strdup(im);
  }else{
	ret->im = NULL;
  }

  uim_prop_list_update(ret->context);

  ret->preedit_head = ret->preedit_tail = NULL;

  ret->cand = (candidate_param *)malloc(sizeof(candidate_param));
  ret->cand->valid = 0;

  return ret;
}




/* add context to context list */
uim_agent_context *
new_uim_agent_context( int id, const char *encoding ){

  uim_agent_context_list *ptr;
  
  m_printf(2, "<FN> add_uim_agent_context(%d)\n", id);

  ptr = (uim_agent_context_list *)malloc(sizeof(uim_agent_context_list));

  ptr->agent_context = create_uim_agent_context(encoding);
  ptr->next = NULL;
  ptr->prev = NULL;

  ptr->agent_context->context_id = id;

  if( agent_context_list_tail != NULL ){
	agent_context_list_tail->next = ptr;
	ptr->prev = agent_context_list_tail;
  }
  agent_context_list_tail = ptr;

  if( agent_context_list_head == NULL )
	agent_context_list_head = ptr;

  return ptr->agent_context;
}


/* release context from context list */
int
release_uim_agent_context( int id ){
  uim_agent_context_list *ptr;

  
  for( ptr = agent_context_list_head; ptr != NULL; ptr = ptr->next ){
	if( ptr->agent_context->context_id == id ){

	  preedit_buffer *p, *ptmp;

	  if( current == ptr->agent_context ){
		current = NULL;
	  }

	  /* release */
	  uim_release_context(ptr->agent_context->context);

	  clear_candidate(ptr->agent_context);
	  free(ptr->agent_context->cand);

	  p = ptr->agent_context->preedit_head; 
	  while( p ){
	  	ptmp = p;
		p = p->next;
		if( ptmp->str )	free( ptmp->str );
		free( ptmp );
	  }
	  free(ptr->agent_context->encoding);

	  /* rebuild list */
	  if( ptr->next != NULL ){
		ptr->next->prev = ptr->prev;
	  }else{
		agent_context_list_tail = ptr->prev;
	  }

	  if( ptr->prev != NULL ){
		ptr->prev->next = ptr->next;
	  }else{
		agent_context_list_head = ptr->next;
	  }

	  free( ptr );

	  return id;
	}
  }

  return -1;
}


/* search context */
uim_agent_context *
get_uim_agent_context( int id ){
  uim_agent_context_list *ptr;

  m_printf(2, "<FN> get_uim_agent_context (%d)\n", id);
  
  for( ptr = agent_context_list_head; ptr != NULL; ptr = ptr->next ){
	if( ptr->agent_context->context_id == id ){
	  return ptr->agent_context;
	}
  }
  return NULL;
}




void
helper_send_im_list( void ){

  int nim, i;

  int len = 0, buflen;
  
  char *buf;

  const char *current_im_name;

  if( current == NULL || current->context == NULL ) return;

  nim = uim_get_nr_im(current->context);

  uim_get_current_im_name(current->context);
  current_im_name = uim_get_current_im_name(current->context);


#define HEADER_FORMAT "im_list\ncharset=%s\n"

  len += strlen(HEADER_FORMAT) + strlen(current->encoding);
  len += strlen("selected") ;

  for(i = 0; i < uim_get_nr_im(current->context); i++){
	const char *name, *lang, *shortd;
	name = uim_get_im_name(current->context, i);
	lang = uim_get_im_language(current->context , i);
	shortd = uim_get_im_short_desc(current->context , i);

	len += name ? strlen(name) : 0;
	len += lang ? strlen(lang) : 0;
	len += shortd ? strlen(shortd) : 0;
	len += strlen( "\t\t\t\n");
  }
  len++ ;

  buf = (char *)malloc(sizeof(char) * len);

  buflen = snprintf(buf, len, HEADER_FORMAT, current->encoding);


  for(i = 0 ; i < uim_get_nr_im(current->context); i++){
	const char *name, *lang, *shortd;	
	name = uim_get_im_name(current->context, i);
	lang = uim_get_im_language(current->context, i);
	shortd = uim_get_im_short_desc(current->context, i);

	m_printf(3, "<FN>  [%d] = %s %s %s\n", i, name, lang, shortd);

	buflen += snprintf(buf + buflen, len - buflen,
					   "%s\t%s\t%s\t%s\n",
					   name ? name : "" ,
					   lang ? lang : "" ,
					   shortd ? shortd : "" ,
					   strcmp(name, (current_im_name == NULL ? "" : current_im_name) ) == 0 ?
					   "selected" : "");
  }

  uim_helper_send_message(helper_fd, buf);

  m_printf(3, "<FN>  im_list = \"%s\"\n", buf);
  
  free(buf);
}



uim_agent_context *
switch_im( uim_agent_context *ua, const char *im, const char *encoding ){

  m_printf(1, "<FN> switch_im\n");

  if( encoding == NULL ){
	/* unspported engine or not ready for use from Emacs  */
	m_printf(1, "<WARNING>  switch_im: im is %s, encoding is NULL\n", im);
	return ua;
  }

  if( strcmp(ua->encoding, encoding) == 0 ){

	if( ua->im != NULL &&
		strcmp(ua->im, im) == 0 ){
	  /* Do Nothing */
	  m_printf(1, "<FN>  same im\n");
	  return ua;
	}else{
	  m_printf(1, "<FN>  different engine but same encoding %s %s\n", ua->im, im);

	  /* uim.el must ignore rest of this line */
	  if(ua->im) free(ua->im);
	  if( im ){
		ua->im = strdup(im);
	  }else{
		ua->im = NULL;
	  }
	  /*
	  ua->im = strdup(uim_get_current_im_name(ua->context));
	  */

	  uim_switch_im(ua->context, im);


	  /* if( im ) ua->im = strdup(im);
		 else ua->im = NULL;*/


	  uim_prop_label_update(ua->context);
	  uim_prop_list_update(ua->context);
	  return ua;
	}
  }else{
	m_printf(1, "<FN>  different encoding %s %s\n", ua->encoding, encoding );

	/* uim.el must ignore rest of this line */

	clear_candidate(ua);
	clear_preedit(ua);
	uim_release_context(ua->context);

	ua->context = create_context(encoding, ua);

	uim_switch_im(ua->context, im);

	if( ua->im ) free(ua->im);
	ua->im = strdup(uim_get_current_im_name(ua->context));
	/* ua->im = strdup(im); */

	if( ua->encoding ) free( ua->encoding );
	ua->encoding = strdup( encoding );

	uim_prop_label_update(ua->context);
	uim_prop_list_update(ua->context);

	return ua;
  }
}



void
helper_im_changed( char *request, char *engine_name ){

  uim_agent_context_list *ptr;

  m_printf(2, "<FN> helper_im_changed: %s\n", engine_name);

  if( strcmp(request, "im_change_this_text_area_only") == 0 ){

    if( current ) {

      switch_im(current, engine_name, get_im_encoding(engine_name));

	  if(current->im) free(current->im);
	  current->im = strdup(engine_name);

      uim_prop_label_update(current->context);
      uim_prop_list_update(current->context);
    }

  }else if( strcmp(request, "im_change_whole_desktop") == 0 ){

	/* change default */
	if( default_engine_name ) free(default_engine_name);
	default_engine_name = strdup(engine_name);

	for( ptr = agent_context_list_head; ptr != NULL; ptr = ptr->next ){

      switch_im(ptr->agent_context, engine_name, get_im_encoding(engine_name));

	  if( ptr->agent_context->im ) free(ptr->agent_context->im);
	  ptr->agent_context->im = strdup(engine_name);

      if( current && ptr->agent_context == current ){
        uim_prop_label_update(ptr->agent_context->context);
        uim_prop_list_update(ptr->agent_context->context);
	  }

	  uim_prop_update_custom(ptr->agent_context->context,
							 "custom-preserved-default-im-name", engine_name);

    }

  }else if( strcmp(request, "im_change_this_application_only") == 0 ){

	/* change default */
	if( default_engine_name ) free(default_engine_name);
	default_engine_name = strdup(engine_name);

	if( current ){

	  for( ptr = agent_context_list_head; ptr != NULL; ptr = ptr->next ){

        switch_im(ptr->agent_context, engine_name, get_im_encoding(engine_name));
		if( ptr->agent_context->im ) free(ptr->agent_context->im);
		ptr->agent_context->im = strdup(engine_name);
		
        if( ptr->agent_context == current ){
		  uim_prop_label_update(ptr->agent_context->context);
          uim_prop_list_update(ptr->agent_context->context);
		}

		uim_prop_update_custom(ptr->agent_context->context,
							   "custom-preserved-default-im-name", engine_name);

      }
	}
  }
}



/* focus in to a buffer */
int
focused( uim_agent_context *ua ){

  m_printf(3, "<FN> focused\n");

  if( ua == NULL || ua->context == NULL ) return -1;

  uim_helper_client_focus_in( ua->context );

  current = ua;

  uim_prop_label_update( ua->context );
  uim_prop_list_update( ua->context );

  return ua->context_id;

}


/* focus out from current buffer */
int
unfocused( void ){

  int ret;

  m_printf(3, "<FN> unfocused\n");

  if( current == NULL || current->context == NULL ) return -1;

  ret = current->context_id;
  uim_helper_client_focus_out( current->context );

  current = NULL;

  return ret;
}



void
helper_handler( void ){

  char *message;

  int current_exist = (current != NULL && current->context != NULL) ? 1 : 0;

  uim_helper_read_proc(helper_fd);

  while( (message = uim_helper_get_message()) ) {

    char *line = message;
    char *eol;

    if( (eol = strchr(line, '\n')) != NULL )
	  *eol = '\0';
    else{
	  free( message );
	  break;
	}

	m_printf(1, "<DEBUG> message \"%s\"\n", message);

	if( strcmp("prop_list_get", line) == 0 ){ /* for current context */


	  /* m_printf(2, "prop_list_get\n"); */
	  if( current_exist ){
		uim_prop_list_update(current->context);
	  }else{
		m_printf(2, "<NOTE> ignored helper message: %s\n", message);
	  }
		
	}else if( strcmp("prop_label_get", line) == 0 ){ /* for current context */

	  /* m_printf(2, "prop_label_get\n"); */
	  if( current_exist ){
		uim_prop_label_update(current->context);
	  }else{
		m_printf(2, "<NOTE> ignored helper message: %s\n", message);
	  }

	}else if( strcmp("prop_activate", line) == 0 ){ /* for current context */

	  /* m_printf(2, "prop_activate\n"); */

	  if( current_exist ){
		line = eol + 1;
		if( (eol = strchr(line, '\n')) != NULL ){
		  *eol = '\0';
		  uim_prop_activate(current->context, line);
		}else{
		  m_printf(1, "<WARNING> invalid message(prop_activate) \"%s\"\n", message);
		}
	  }else{
		m_printf(2, "<NOTE> ignored helper message: %s\n", message);
	  }

	}else if( strcmp("im_list_get", line) == 0 ){ /* for current context */

	  if( current_exist ){
		/* m_printf(2, "im_list_get\n"); */
		helper_send_im_list();
	  }else{
		m_printf(2, "<NOTE> ignored helper message: %s\n", message);
	  }
		
	}else if( strncmp("focus_in", line, 8) == 0 ){

	  if( current_exist ){
		/* m_printf(2, "focus_in\n"); */
		if( current ) unfocused();
	  }else{
		m_printf(2, "<NOTE> ignored helper message: %s\n", message);
	  }

	}else if( strcmp("commit_string", line) == 0 ){ /* for current context */

	  if( current_exist ){
		/* m_printf(2, "commit_string\n"); */
		line = eol + 1;
		if( (eol = strchr(line, '\n')) != NULL ){
		  *eol = '\0';
		  /*   '\' and '"' should be escaped */
		  /* a_printf("( -1 ( s \"%s\" ) ( e ) )\n", line); */
		}else{
		  m_printf(1, "<WARNING> invalid message(commit_string) \"%s\"\n", message);
		}
	  }else{
		m_printf(2, "<NOTE> ignored helper message: %s\n", message);
	  }

	}else if( strncmp(line, "im_change_", 10) == 0 ){

	  char *engine;

	  engine = eol + 1;
	  if( (eol = strchr(engine, '\n')) != NULL ){
		*eol = '\0';
		/* m_printf("line %s\nengine: %s\n", line, engine); */
		helper_im_changed(line, engine);
	  }else{
		m_printf(1, "<WARNING> invalid message(im_change) \"%s\"\n", message);
	  }

	}else if( strcmp("prop_update_custom", line) == 0 ){

	  /* message from uim-pref-{gtk,qt} */
	  char *custom;
	  char *val;

	  uim_agent_context_list *ptr;

	  /* m_printf("prop_update_custom\n"); */

	  line = eol + 1;
	  custom = line;
	  if( (eol = strchr(line, '\n')) != NULL ){
		*eol = '\0';

		val = eol + 1;

		if( (eol = strchr(val, '\n')) != NULL ){
		  *eol = '\0';

		  /* m_printf("%s : %s\n", val, custom); */
		  for( ptr = agent_context_list_head; ptr != NULL; ptr = ptr->next ){
			uim_prop_update_custom(ptr->agent_context->context, custom, val);
		  }

		}else{
		  m_printf(1, "<WARNING> invalid message val of prop_update_custom: \"%s\"\n", val);
		}

	  }else{
		m_printf(1, "<WARNING> invalid message: custom of prop_update_custom: \"%s\"\n", line);
	  }

	}else if( strcmp("prop_list_update", line) == 0 ){

	  /* ignore */
	  /* m_printf("prop_list_update (helper)\n"); */

	}else if( strcmp("prop_label_update", line) == 0 ){

	  /* ignore */
	  /* m_printf("prop_label_update (helper)\n"); */

	}else{
	  m_printf(1, "<WARNING> undefined helper message: %s\n", message);
	}

	free(message);
  }

}



/* called when helper_fd is disconnected */
void
helper_disconnected_cb( void ){
  m_printf(1, "<WARNING> helper_disconnected\n");
  helper_fd = -1;
}



void
check_helper_connection( void ){
  if( helper_fd < 0 ){
    helper_fd = uim_helper_init_client_fd(helper_disconnected_cb);
  }
}







static int
cmd_release( int serial, int context_id ){
  
  a_printf(" ( %d 0 ", serial);

  if( release_uim_agent_context(context_id) < 0 ){
	/*m_printf(3, "release_uim_agent_context error\n"); */
	a_printf(" ( f ) )\n");
	return -1;
  }else{
	/* m_printf(3, "release_uim_agent_context OK\n"); */
	a_printf(" ( a ) )\n");
	return 1;
  }
}


static int
cmd_unfocused( int serial, int context_id ){

  int ret;
  uim_agent_context *ua;

  /* m_printf("UNFOCUSED\n"); */

  /* if context_id is 0, 
	 this function unfocuses from current focused context 
	 otherwise,
	 compare current focused context and unfocuses if they are matching.
  */

  ua = current;

  if( current == NULL ){
	a_printf("( %d 0 ( f ) )\n", serial);
	return -1;
  }

  if( context_id != 0 && get_uim_agent_context(context_id) != current ){
	a_printf("( %d 0 ( f ) )\n", serial);
	return -1;
  }

  ret = unfocused();

  if( ret > 0 ){
	if( context_id != 0 ){
	  a_printf("( %d %d ", serial, context_id);
	  show_preedit(ua);
	  a_printf(" ) ( a ) )\n");
	  return 1;
	}else{
	  a_printf("( %d %d ( a ) )\n", serial, context_id);
	  return 1;
	}
  }else{
	a_printf("( %d 0 ( f ) )\n", serial);
	return -1;
  }

}


static int
cmd_focused( int serial, int context_id ){

  uim_agent_context *ua;

  a_printf("( %d %d ", serial, context_id);

  /*m_printf("FOCUSED\n"); */
  if( focused(ua = get_uim_agent_context(context_id)) > 0){
	if( show_preedit(ua) ){
	  show_candidate(ua);
	}
	a_printf(" ( a ) )\n");
	return 1;
  }else{
	a_printf(" ( f ) )\n");
	return -1;
  }
}

static int
cmd_hide( int serial, int context_id ){

  uim_agent_context *ua;

  a_printf("( %d %d ", serial, context_id);
  
  if( (ua = get_uim_agent_context(context_id)) != NULL ){
	/* hide candidate list (show preedit only) */
	/* show_preedit(ua); */
	a_printf(" ( a ) )\n");
	return 1;
  }else{
	a_printf(" ( f ) )\n");
	return -1;
  }

}

static int
cmd_show( int serial, int context_id ){

  uim_agent_context *ua;
		
  a_printf("( %d %d ", serial, context_id);
  
  if( (ua = get_uim_agent_context(context_id)) != NULL ){
	if( show_preedit(ua) ){
	  show_candidate(ua);
	}
	a_printf(" ( a ) )\n");
	return 1;
  }else{
	a_printf(" ( f ) )\n");
	return -1;
  }
}


static int
cmd_new( int serial, int context_id, const char *encoding ){

  a_printf("( %d 0 ", serial);

  if( get_uim_agent_context(context_id) ){
	m_printf(1, "<WARNING> context %d already exists\n", context_id);
	a_printf(" ( f ) )\n");
	return -1;
  }else{
	if( new_uim_agent_context(context_id, encoding) != NULL ){
	  a_printf(" ( a ) )\n");
	  return 1;
	}else{
	  a_printf(" ( f ) )\n");
	  return -1;
	}
  }
}


static int
cmd_reset( int serial, int context_id ){

  /* reset current context */

  uim_agent_context *ua;


  a_printf("( %d 0 ", serial);

  if( (ua = get_uim_agent_context(context_id)) != NULL ){
	/* before reset, clear preedit and candidate */
	clear_preedit(ua);
	clear_candidate(ua);
	uim_reset_context(ua->context);
	a_printf(" ( a ) )\n");
	return 1;
  }else{
	a_printf(" ( f ) )\n");
	return -1;
  }

}


static int
cmd_change( int serial, int context_id, const char *im, const char *encoding ){

  uim_agent_context *ua;

  a_printf("( %d 0 ", serial);

  if( im && encoding && strlen(im) > 0 
	  && (ua = get_uim_agent_context(context_id)) ){

	focused(ua);

	if( check_im_name(im) ){

	  switch_im(ua, im, encoding);

	  a_printf(" ( a ) )\n");
	  return 1;
	}else{
	  a_printf(" ( f ) )\n");
	  return -1;	 
 	}
  }else{
	a_printf(" ( f ) )\n");
	return -1;
  }

}

static int
cmd_list( int serial ){

  a_printf(" ( %d 0 ", serial);

  if( list_im_engine() > 0 ){
	a_printf(" ( a ) )\n");
	return 1;
  }else{
	a_printf(" ( f ) )\n");
	return -1;
  }
}


static int
cmd_default( int serial, const char *im ){

  m_printf(1, "<default> im = %s\n", im);

  if( default_engine_name ) free(default_engine_name);

  if( im ) default_engine_name = strdup(im);
  else default_engine_name = NULL;

  a_printf(" ( %d 0 ( a ) )\n", serial);

  return 1;
}



static int
cmd_setenc( int serial, const char *im, const char *encoding ){

  a_printf(" ( %d 0 ", serial);

  if( set_im_encoding(im, encoding) > 0 ){
	a_printf(" ( a ) )\n");
	return 1;
  }else{
	a_printf(" ( f ) )\n");
	return -1;
  }
}


static int
cmd_getenc( int serial, const char *im ){

  const char *encoding;

  encoding = get_im_encoding(im);

  a_printf("( %d 0 ", serial);
  if( encoding ){
	a_printf(" ( E \"%s\" ) ( a ) )\n", encoding);
  }else{
	a_printf(" ( E \"\" ) ( f ) )\n");
  }

  return 1;
}


static int
cmd_prop( int serial, int context_id, const char *prop ){

  uim_agent_context *ua;

  a_printf(" ( %d 0 ", serial);

  if( prop && strlen(prop) > 0 &&
	  (ua = get_uim_agent_context(context_id) )){
	uim_prop_activate(ua->context, prop);
	a_printf(" ( a ) )\n");
	return 1;
  }else{
	a_printf(" ( f ) )\n");
	return -1;
  }
}

static int
cmd_error( int serial ){
  a_printf("( %d 0 ( f ) )\n", serial);
  return -1;
}







int
process_command( int serial, int cid, char *cmd ){
  char *opt, *opt2;
  
  if( (opt = strchr(cmd, ' ')) ) {
	*opt = '\0';
	opt ++;
	/*m_printf("opt = %s\n", opt); */

	if( (opt2 = strchr(opt, ' ')) ) {
	  *opt2 = '\0';
	  opt2 ++;
	  /*m_printf("opt = %s\n", opt); */
	}else{
	  opt2 = NULL;
	}

  }else{
	opt = NULL;
	opt2 = NULL;
  }


  if( strcmp(cmd, "RELEASE") == 0 ){
	return cmd_release(serial, cid);
  }else if( strcmp(cmd, "UNFOCUSED") == 0 ){
	return cmd_unfocused(serial, cid);
  }else if( strcmp(cmd, "FOCUSED") == 0 ){
	return cmd_focused(serial, cid);
  }else if( strcmp(cmd, "HIDE") == 0 ){
	return cmd_hide(serial, cid);
  }else if( strcmp(cmd, "SHOW") == 0 ){
	return cmd_show(serial, cid);
  }else if( strcmp(cmd, "NEW") == 0 ){
  	return cmd_new(serial, cid, opt);
  }else if( strcmp(cmd, "RESET") == 0 ){
	return cmd_reset(serial, cid);
  }else if( strcmp(cmd, "CHANGE") == 0 ){
  	return cmd_change(serial, cid, opt, opt2);
  }else if( strcmp(cmd, "LIST") == 0 ){
  	return cmd_list(serial);
  }else if( strcmp(cmd, "DEFAULT") == 0 ){
  	return cmd_default(serial, opt);
  }else if( strcmp(cmd, "SETENC") == 0 ){
  	return cmd_setenc(serial, opt, opt2);
  }else if( strcmp(cmd, "GETENC") == 0 ){
  	return cmd_getenc(serial, opt); /* for debug? */
  }else if( strcmp(cmd, "PROP") == 0 ){
  	return cmd_prop(serial, cid, opt);
  }else{
	return cmd_error(serial);
  }
}



/* [  ] */
int
analyze_keyvector( char *vector, uim_key *ukey, char *keyname ){
  char *p1, *p2, *c;
  int key;

  p1 = vector + 1;

  /* m_printf("analyze_key %s\n", vector); */

  /* check space */
  if( (p2 = strchr(vector, ' ')) ){
	/* if there is a space, 1st string must be 27 */
	key = strtol(p1, &c, 10);
	if( c != p2 || key != 27 ){
	  m_printf(1, "<WARNING> 1st key is not 27\n");
	  return -1;
	}else{
	  /* m_printf("Meta detected\n"); */
	  ukey->mod |= UMod_Alt;
	}
	p1 = p2 + 1; 
  }


  if( *p1 == ']'){
	m_printf(1, "<WARNING> invalid key vector\n");
	return -1;
  }

  key = strtol(p1, &c, 10);

  if( *c == ']' ){
	/* all of rest characters are digit */
	if(key & (1<<27)) ukey->mod |= UMod_Alt;
	/* if(key & (1<<27)) mod |= UMod_Meta; */
	if(key & (1<<26)) ukey->mod |= UMod_Control;
	if(key & (1<<25)) ukey->mod |= UMod_Shift;
	if(key & (1<<24)) ukey->mod |= UMod_Hyper;
	if(key & (1<<23)) ukey->mod |= UMod_Super;
	if(key & (1<<22)) ukey->mod |= UMod_Alt;
	ukey->key = key & 0xff;

	if( ukey->key == 27 ){
	  ukey->key = UKey_Escape;
	}else if( ukey->key <= 26 ){
	  ukey->mod |= UMod_Control;
	  if( ukey->key == 0 ){
		ukey->key = 32;
	  }else{
		ukey->key += 96;
	  }
	}
  }else{
	while( 1 ){
	  if( *(p1 + 1) == '-' ){
		switch(*p1){
		case 'S':
		  ukey->mod |= UMod_Shift;
		  break;
		case 'C':
		  ukey->mod |= UMod_Control;
		  break;
		case 'A':
		  ukey->mod |= UMod_Alt;
		  break;
		case 'M':
		  /* mod |= UMod_Meta; */
		  ukey->mod |= UMod_Alt;
		  break;
		case 's':
		  ukey->mod |= UMod_Super;
		  break;
		case 'H':
		  ukey->mod |= UMod_Hyper;
		  break;
		case 'X':
		  /* dummy; */
		  break;
		default:
		  return -1;
		  break;
		}
		p1 += 2;
	  }else{
		if( (p2 = strrchr(p1, ']')) == NULL || p1 == p2 ){
		  /* [C-] [C-M-] */
		  m_printf(1, "<WARNING> invalid key vector\n");
		  return -1;
		}else{
		  *p2 = '\0';

		  if( strlen(p1) == 1 ){
			/* single character [C-a] [C-1] */
			ukey->key = *p1;
		  }else{
			/* [C-next] [C-M-end] */
			if( *p1 >= 'A' && *p1 <= 'Z' ){
			  convert_keyname(keyname, p1);
			}else{
			  strcpy(keyname, p1);
			}
			ukey->key = matchkey( keyname );
		  }
		  break;
		}
	  }
	}
  }

  if( ukey->key < 0 ){
	return -1;
  }
  
  return 1;

}


int
process_keyvector( int serial, int cid, uim_key ukey, const char *keyname ){

  int ret;

  a_printf("( %d %d ", serial, cid);

  if( current == NULL || 
	  (current != NULL && current->context_id != cid) ){

	if( focused(get_uim_agent_context(cid)) < 0 ){
	  m_printf(1, "<WARNING> context %d not found\n", cid);
	  a_printf(" ( f ) )\n");
	  return -1;
	}

  }


  /* key input is received by requested context */
  ret = uim_press_key(current->context, ukey.key, ukey.mod);
  uim_release_key(current->context , ukey.key, ukey.mod);

  if( ret > 0 ){
	/* uim did not process the key */
	if( current->preedit_head == NULL || current->preedit_length == 0 ){
	  /* preedit doesn't exist */
	  if( ukey.mod != 0 || ukey.key > 255 ){

		a_printf(" ( n [(");
		if( ukey.mod & UMod_Control ) a_printf("control ");
		if( ukey.mod & UMod_Alt ) a_printf("meta ");
		/* if( ukey->mod & UMod_Shift ) a_printf("shift "); */
		if( ukey.mod & UMod_Hyper ) a_printf("hyper ");
		if( ukey.mod & UMod_Super ) a_printf("super ");
		
		if( ukey.key > 255 )
		  a_printf("%s", keyname);
		else
		  a_printf("%d", ukey.key);
		
		a_printf(")] ) ");

	  }else{
		a_printf(" ( n [%d] ) ", ukey.key);
	  }
			
	}else{
	  /* discard the key if preedit exists */
	  show_preedit(current);
	  show_candidate(current);
	  a_printf(" ( n ) ");
	}
			
  }else{
	/* uim process the key */
	show_preedit(current);
	show_candidate(current);
  }

  a_printf(" ) ");
  a_printf("\n");

  return 1;
}




void
wait_until_data_arrival( fd_set *rfds ){
  int fdmax;

  fdmax = STDIN_FILENO;

  if( helper_fd > 0){
	FD_ZERO(rfds);
	FD_SET(helper_fd, rfds);
	if( helper_fd > fdmax  ) fdmax = helper_fd;
  }

  FD_SET(STDIN_FILENO, rfds);
  
  if( select(fdmax + 1, rfds, NULL, NULL, NULL) < 0 ){
	m_printf(1, "<ERROR> select error\n");
  }
}


int
take_one_line_from_buffer( char *buf, char *input ){
  char *p = strchr(buf, '\n'); 
  int rest; 

  if( !p ) {
	m_printf(1, "<ERROR> something wrong occured\n");
	return 0;
  }

  *p = '\0';
  strcpy(input, buf);

  rest = strlen(p + 1);
  memmove(buf, (p + 1), rest);

  return rest;
}




int
main( int argc, char *argv[] ){

  int opt;

  while( (opt = getopt( argc, argv, "d" )) != -1 ){
	switch(opt){
	case 'd':
	  debug ++;
	  break;
	}
  }

  /*  if( argc > 1 ){
	if( !strcmp(argv[1], "-d") ){
	  debug = 1;
	}
	}*/

  if( ! debug ) fclose(stderr);

#ifdef UIM_AGENT_DEBUG
  if( !debug ){
	if( (debugf = fopen("uim-agent.log", "w")) == NULL ){
	  perror("uim-agent.log");
	}
  }
#endif

  if( uim_init() < 0 ){
	fprintf(stderr, "uim_init error\n");
	return -1;
  }



  while(1){
	int cid, serial;
	char *p1, *p2, *c;
	int offset = 0, more_input = 0;
	char input[256], buf[256], keyname[32];

	uim_key ukey;
	fd_set rfds;

	fflush(stdout);

	buf[sizeof(buf) - 1] = '\0';

	if( ! more_input ){

	  check_helper_connection();

	  wait_until_data_arrival( &rfds );

	  if( FD_ISSET(STDIN_FILENO, &rfds) ){
		/* read from STDIN if input data exists  */
		read(STDIN_FILENO, buf + offset, sizeof(buf) - offset - 1);
	  }else if( FD_ISSET(helper_fd, &rfds) ){
		/* process helper message only when input data doesn't exist */
		output_enable = 0;
		helper_handler();
		output_enable = 1;
	  }
	}


	if( (p1 = strchr(buf, '\n')) ){
	  offset = take_one_line_from_buffer( buf, input );
	  more_input = strchr(buf, '\n') ? 1 : 0;
	}else{
	  /* wait \n */
	  offset = strlen(buf);
	  more_input = 0;
	  continue;
	}


	p1 = input;
	serial = -1;

	/*
	  command format 
	    serial CID COMMAND OPTION

	  key format
	    serial CID [keyvector]
	*/

	if( (p2 = strchr(p1, ' ')) == NULL ){
	  m_printf(1, "<WARNING> input error: space after 1st string\n");
	  goto ERROR;
	}

	/* 1st string must be digit */
	*p2 = '\0';
	serial = strtol(p1, &c, 10);
	if( c != p2 ){
	  m_printf(1, "<WARNING> input error: invalid serial %d\n", serial);
	  goto ERROR;
	}

	p1 = p2 + 1;
	if( (p2 = strchr(p1, ' ')) == NULL ){
	  m_printf(1, "<WARNING> input error: no space after 2nd string\n");
	  goto ERROR;
	}

	/* 2nd string must be digit */
	*p2 = '\0';
	cid = strtol(p1, &c, 10);
	if( c != p2 ){
	  m_printf(1, "<WARNING> invalid cid %d\n", cid);
	  goto ERROR;
	}

	/* 3rd string */
	p1 = p2 + 1;

	if( *p1 == '[' ){
	  /* keyvector if 3rd string starts with [  */
	  /*m_printf(3, "key vector\n");*/

	  if( (p2 = strchr(p1, ']')) == NULL) {
		/* no corresponding ]  */
		m_printf(1, "<WARNING> ']' not found\n");
		goto ERROR;
	  }


	  p2 ++; 
	  if( *p2 == ']') p2 ++; /* for [X-]] */
	  *p2 = '\0';   /* replace character after ] with \0  */

	  ukey.mod = 0;
	  ukey.key = -1;
	  keyname[0] = '\0';

	  /*m_printf("***analyze vector\n");*/
	  if( analyze_keyvector(p1, &ukey, keyname) > 0 ){
		process_keyvector(serial, cid, ukey, keyname);
		continue;
	  }

	  goto ERROR;


	}else if( *p1 >= 'A' && *p1 <= 'Z' ){
	  /* command */
	  if( process_command(serial, cid, p1) < 0 ){
		m_printf(1, "<WARNING> command error\n");
	  }
	  continue;
	}
  
	m_printf(1, "<WARNING> invalid input\n");
	
  ERROR:
	a_printf("( %d 0 ( x ) )\n", serial);
  }

  uim_quit();
  return 0;

}

